#ifndef _LIGHT_H_
#define _LIGHT_H_
#include "vector.h"
#include <math.h>

const double DEG = 180 / M_PI;

class Light {
  protected:
    Vector color;
  public:
    Vector org;
    Light(Vector org, Vector color)
    {
      Light::org = org;
      Light::color = color;
    }
    virtual Vector GetColor(double, Vector, Vector) = 0;
    Light *next;
};

class PointLight: public Light {
  public:
    PointLight(Vector org, Vector color): Light(org, color) { }
    Vector GetColor(double density, Vector dir, Vector n)
    {
      return color * exp(-density * (!dir)) * fabs(Normalize(dir) & n);
    }
};

class SpotLight: public PointLight {
    Vector dir;
    double angle, blur;
  public:
    SpotLight(Vector org, Vector color, Vector d, double a, double blur):
    PointLight(org, color), dir(d), angle(a)
    {
      SpotLight::blur = blur;
    }
    Vector GetColor(double density, Vector d, Vector n)
    {
      double a = acos((dir & -d) / (!dir * !d)) * DEG;
      if (a < angle)
      {
        Vector c = PointLight::GetColor(density, d, n);
        if (a < angle - blur)
          return c;
        else
          return c * (1 - (a - angle + blur) / blur);
      }
      return 0;
    }
};

#endif
