#include <windows.h>
#include <fstream.h>
#include <string.h>
#include <stdio.h>
#include "scene.h"

#define WIDTH        640
#define HEIGHT       480
#define CMD_RENDER   101
#define CMD_SAVEAS   102
#define CMD_EXIT     103
#define CMD_ABOUT    104
#define MS_COMPLETED 1001
#define MINRES       1
#define MAXRES       10000

const int widths[] = {320, 320, 512, 640, 640, 800, 1024, 1280, 1600};
const int heights[] = {200, 240, 384, 400, 480, 600, 768, 1024, 1200};

struct {
  int res;
  int x_res;
  int y_res;
  bool save;
  char filename[256];
} dlg_struct = {4, 640, 480, BST_UNCHECKED, "noname.bmp"}, temp_struct;

struct ThreadStruct{
  int width;
  int height;
  int bps;
  HWND hWnd;
} ts = {640, 480, 0, 0};

char header[] = {0x42,0x4D,0x36,0x84,0x03,0x00,0x00,0x00,0x00,0x00,0x36,0x00,0x00,0x00,0x28,0x00,
                 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x18,0x00,0x00,0x00,
                 0x00,0x00,0x00,0x84,0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                 0x00,0x00,0x00,0x00,0x00,0x00};

HANDLE RenderThread;
HINSTANCE hInstance;
HWND hDlg;
char *render_buf = NULL;
unsigned long ret_val;
bool draw_buffer = false;
Scene *scene;

SYSTEMTIME SubtractTime(SYSTEMTIME, SYSTEMTIME);
DWORD WINAPI ThreadFunc(LPVOID);
void ResetRenderer();
void CreateBuffers();
void CreateScene();
bool TestString(char *, int&);
bool GetFileName(char *, HWND);
void SaveFile(char *);
BOOL WINAPI dlg1Proc(HWND, UINT, WPARAM, LPARAM);
BOOL WINAPI dlg2Proc(HWND, UINT, WPARAM, LPARAM);
BOOL WINAPI dlg3Proc(HWND, UINT, WPARAM, LPARAM);
void SetScrollBars(int, int, HWND);
void ConvertColor(long, long, long, long, long, long, int, char *);
LONG WINAPI wndProc(HWND, UINT, WPARAM, LPARAM);

SYSTEMTIME SubtractTime(SYSTEMTIME t1, SYSTEMTIME t2)
{
  SYSTEMTIME t;
  int s, m, h;
  h = t1.wHour - t2.wHour;
  m = t1.wMinute - t2.wMinute;
  s = t1.wSecond - t2.wSecond;
  if (s < 0)
  {
    s += 60;
    m--;
  }
  if (m < 0)
  {
    m += 60;
    h--;
  }
  t.wHour = h;
  t.wMinute = m;
  t.wSecond = s;
  return t;
}

DWORD WINAPI ThreadFunc(LPVOID)
{
  SYSTEMTIME t1, t2, t3;
  char t[100];

  GetSystemTime(&t1);
  scene -> RenderScene(ts.width, ts.height, 0.64, 0.48, Vector(0, 0, 2), Vector(1, 0, 0), Vector(0, 0, 1), 32, render_buf);
  GetSystemTime(&t2);
  SendMessage(ts.hWnd, MS_COMPLETED, 0, 0);
  t3 = SubtractTime(t2, t1);
  sprintf(t, "Start time: %d:%02d:%02d\nEnd time: %d:%02d:%02d\nElapsed time: %2d:%02d:%02d",
          t1.wHour, t1.wMinute, t1.wSecond,
          t2.wHour, t2.wMinute, t2.wSecond,
          t3.wHour, t3.wMinute, t3.wSecond);
  MessageBox(ts.hWnd, t, "Information", MB_OK + MB_ICONINFORMATION);

  ExitThread(0);
  return 0;
}

void ResetRenderer()
{
  TerminateThread(RenderThread, 0);
  CloseHandle(RenderThread);
}

void CreateBuffers()
{
  long size = ts.width * ts.height * 3;
  if (render_buf != NULL)
    delete render_buf;
  render_buf = new char[size];
  ZeroMemory(render_buf, size);
}

void CreateScene()
{
  SurfaceData data;
  data.Ka = 0;
  data.Kd = 1;
  data.Kr = 0;
  data.Ks = 0;
  data.Kt = 0;
  data.p = 0;
  data.med = Medium(1.02, 0.05);
  scene = new Scene(Vector(0.8, 0.8, 1.0), Vector(0), Medium(1, 0.05));
  scene -> AddObject(new XYPlane(data, 0, new Texture("tex3.bmp", 0, 0, 0.05)));
  data.Kd = 0.3;
  data.Ks = 0.3;
  data.Kr = 0.4;
  data.p = 4;
  scene -> AddObject(new Sphere(data, Vector(10, -2, 2), 2, Vector(1, 1, 0)));
  data.Kd = 0.2;
  data.Ks = 0.5;
  data.Kr = 0.2;
  data.Kt = 0.5;
  scene -> AddObject(new Sphere(data, Vector(7, 3, 1), 1, Vector(0, 0, 1)));
  data.Kt = 0;
  data.Kd = 1;
  data.Ks = 0;
  data.Kr = 0;
  scene -> AddObject(new Sphere(data, Vector(14, 3, 1.5), 1.5, new Texture("tex6.bmp", 0, 0, 1)));
  data.Kd = 1;
  data.Kr = 0;
  data.Ks = 0;
  scene -> AddObject(new Cube(data, Vector(5, 1, 0.5), 0.5, 0.5, 0.5, new Texture("tex5.bmp", 0, 0, (double)1 / 128)));
  data.Kd = 0;
  data.Ks = 0;
  data.Kr = 1;
  scene -> AddObject(new Sphere(data, Vector(6, -3, 0.5), 0.5, Vector(1)));
  data.Kd = 0.8;
  data.Kr = 0.2;
  data.Ks = 0;
  data.p = 0;
  scene -> AddObject(new XYPlane(data, 6, new Texture("tex4.bmp", 0, 0, 0.05)));
  data.Kd = 1;
  data.Kr = 0;
  scene -> AddObject(new YZPlane(data, 18, new Texture("tex2.bmp", 0, 0, 0.05)));
  scene -> AddObject(new YZPlane(data, -2, new Texture("tex2.bmp", 0, 0, 0.05)));
  scene -> AddObject(new ZXPlane(data, 6, new Texture("tex1.bmp", 0, 0, 0.05)));
  scene -> AddObject(new ZXPlane(data, -6, new Texture("tex1.bmp", 0, 0, 0.05)));
  scene -> AddLight(new PointLight(Vector(15, 3, 5), Vector(1)));
  scene -> AddLight(new SpotLight(Vector(6, -3, 2), Vector(1, 1, 0), Vector(0, 0, -1), 45, 10));
  scene -> AddLight(new PointLight(Vector(0, 0, 3), Vector(0.7, 0.7, 1)));
  scene -> SetFog(-0.1, 1.5, Vector(0.8, 0.8, 1.0), 0.1, 0.05, 3);
  scene -> EnableFog();
}

bool TestString(char *s, int& v)
{
  if (!strlen(s)) return false;

  v = atoi(s);
  if (v < MINRES || v > MAXRES) return false;

  return true;
}

bool GetFileName(char *buf, HWND owner)
{
  OPENFILENAME o;

  o.lStructSize = sizeof(OPENFILENAME);
  o.hwndOwner = owner;
  o.hInstance = 0;
  o.lpstrFilter = "Windows bitmap (*.BMP)\x0*.BMP\x0\x41ll files (*.*)\x0*.*\x0\x0";
  o.lpstrCustomFilter = NULL;
  o.nMaxCustFilter = 0;
  o.nFilterIndex = 0;
  o.lpstrFile = buf;
  o.nMaxFile = 256;
  o.lpstrFileTitle = NULL;
  o.nMaxFileTitle = 0;
  o.lpstrInitialDir = NULL;
  o.lpstrTitle = "Save...";
  o.Flags = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;
  o.nFileOffset = 0;
  o.nFileExtension = 0;
  o.lpstrDefExt = "BMP";
  o.lCustData = 0;
  o.lpfnHook = NULL;
  o.lpTemplateName = NULL;

  return GetSaveFileName(&o);
}

void SaveFile(char *filename)
{
  ofstream out(filename, ios::binary);
  if (!out)
  {
    char t[100] = "Cannot create file:\n";
    strcat(t, filename);
    MessageBox(0, t, "Error", MB_OK + MB_ICONSTOP);
    return;
  }
  *(long *)(header + 0x12) = ts.width;
  *(long *)(header + 0x16) = ts.height;
  out.write(header, sizeof(header));
  for (int i = 0; i < ts.height; i++)
    out.write(render_buf + ts.width * (ts.height - i - 1) * 3, ts.width * 3);
  out.close();
}

BOOL WINAPI dlg1Proc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  char t[6], t2[6], buf[256];
  HWND edit1 = GetDlgItem(hWnd, 112),
       edit2 = GetDlgItem(hWnd, 113),
       edit3 = GetDlgItem(hWnd, 116);

  switch (msg)
  {
    case WM_INITDIALOG:
      CheckRadioButton(hWnd, 102, 111, dlg_struct.res + 102);
      sprintf(t, "%d", dlg_struct.x_res);
      SendMessage(edit1, EM_REPLACESEL, FALSE, (LONG)t);
      sprintf(t, "%d", dlg_struct.y_res);
      SendMessage(edit2, EM_REPLACESEL, FALSE, (LONG)t);
      SendMessage(edit1, EM_SETLIMITTEXT, 5, 0);
      SendMessage(edit2, EM_SETLIMITTEXT, 5, 0);
      SendMessage(edit3, EM_SETLIMITTEXT, 255, 0);
      CheckDlgButton(hWnd, 115, dlg_struct.save);
      SetDlgItemText(hWnd, 116, dlg_struct.filename);
      if (dlg_struct.res != 9)
      {
        EnableWindow(edit1, FALSE);
        EnableWindow(edit2, FALSE);
      }
      if (!dlg_struct.save)
      {
        EnableWindow(edit3, FALSE);
        EnableWindow(GetDlgItem(hWnd, 117), FALSE);
      }
      return true;

    case WM_COMMAND:
      if (LOWORD(wParam) >= 102 && LOWORD(wParam) <= 111 && HIWORD(wParam) == BN_CLICKED)
      {
        if (LOWORD(wParam) != 111)
        {
          EnableWindow(edit1, FALSE);
          EnableWindow(edit2, FALSE);
        }
        else
        {
          EnableWindow(edit1, TRUE);
          EnableWindow(edit2, TRUE);
        }
        temp_struct.res = LOWORD(wParam) - 102;
      }
      switch (LOWORD(wParam))
      {
        case 115:
          if (HIWORD(wParam) != BN_CLICKED) break;
          temp_struct.save = IsDlgButtonChecked(hWnd, 115);
          if (temp_struct.save == BST_CHECKED)
          {
            EnableWindow(edit3, TRUE);
            EnableWindow(GetDlgItem(hWnd, 117), TRUE);
          }
          else
          {
            EnableWindow(edit3, FALSE);
            EnableWindow(GetDlgItem(hWnd, 117), FALSE);
          }
          break;

        case 117:
          if (HIWORD(wParam) != BN_CLICKED) break;
          GetDlgItemText(hWnd, 116, temp_struct.filename, 256);
          strcpy(buf, temp_struct.filename);
          if (GetFileName(buf, hWnd))
          {
            strcpy(temp_struct.filename, buf);
            SetDlgItemText(hWnd, 116, temp_struct.filename);
          }
          break;

        case IDOK:
          if (temp_struct.res == 9)
          {
            GetDlgItemText(hWnd, 112, t, 6);
            GetDlgItemText(hWnd, 113, t2, 6);
            if (!TestString(t, temp_struct.x_res) || !TestString(t2, temp_struct.y_res))
            {
              MessageBox(hWnd, "Resolution must be between 1 x 1 and 10000 x 10000", "Error", MB_OK + MB_ICONSTOP);
              return false;
            }
          }
          GetDlgItemText(hWnd, 116, temp_struct.filename, 256);
          dlg_struct = temp_struct;
          EndDialog(hWnd, IDOK);
          break;

        case IDCANCEL:
          EndDialog(hWnd, IDCANCEL);
      }
      return false;
  }
  return false;
}

BOOL WINAPI dlg2Proc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  char t[40];
  static SYSTEMTIME time;
  SYSTEMTIME time2, time3;

  switch (msg)
  {
    case WM_INITDIALOG:
      EnableMenuItem(GetMenu(ts.hWnd), CMD_RENDER, MF_GRAYED);
      EnableMenuItem(GetMenu(ts.hWnd), CMD_SAVEAS, MF_GRAYED);
      GetSystemTime(&time);
      sprintf(t, "Start time: %d:%02d:%02d", time.wHour, time.wMinute, time.wSecond);
      SetDlgItemText(hWnd, 200, t);
      sprintf(t, "Current time: %d:%02d:%02d", time.wHour, time.wMinute, time.wSecond);
      SetDlgItemText(hWnd, 201, t);
      sprintf(t, "Resolution: %d x %d", ts.width, ts.height);
      SetDlgItemText(hWnd, 203, t);
      sprintf(t, "Memory used: %.2fMb", (ts.width * ts.height * 3 + ts.height * (ts.width & 1)) / 1048576.0);
      SetDlgItemText(hWnd, 204, t);
      SetTimer(hWnd, 1, 1000, NULL);
      return true;

    case WM_COMMAND:
      switch (LOWORD(wParam))
      {
        case 101:
          ResetRenderer();
          DestroyWindow(hWnd);
          break;
        case 102:
          GetDlgItemText(hWnd, 102, t, 20);
          if (t[0] == 'P')
          {
            SuspendThread(RenderThread);
            InvalidateRect(ts.hWnd, NULL, FALSE);
            SetDlgItemText(hWnd, 102, "Resume rendering");
          }
          else
          {
            ResumeThread(RenderThread);
            SetDlgItemText(hWnd, 102, "Pause rendering");
          }
          break;
      }
      return false;

    case WM_TIMER:
      GetSystemTime(&time2);
      sprintf(t, "Current time: %d:%02d:%02d", time2.wHour, time2.wMinute, time2.wSecond);
      SetDlgItemText(hWnd, 201, t);
      time3 = SubtractTime(time2, time);
      sprintf(t, "Elapsed time: %d:%02d:%02d", time3.wHour, time3.wMinute, time3.wSecond);
      SetDlgItemText(hWnd, 202, t);
      sprintf(t, "Completed: %.2f%", scene -> GetCompletedVal());
      InvalidateRect(ts.hWnd, NULL, FALSE);
      SetDlgItemText(hWnd, 205, t);
      break;

    case WM_DESTROY:
      EnableMenuItem(GetMenu(ts.hWnd), CMD_RENDER, MF_ENABLED);
      KillTimer(hWnd, 1);
      EndDialog(hWnd, 0);
  }
  return false;
}

BOOL WINAPI dlg3Proc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  PAINTSTRUCT ps;
  HDC DC, MemDC;
  HBITMAP bm;

  switch (msg)
  {
    case WM_INITDIALOG: return true;
    case WM_COMMAND:
      if (LOWORD(wParam) == IDOK) EndDialog(hWnd, IDOK);
      if (LOWORD(wParam) == IDCANCEL) EndDialog(hWnd, IDCANCEL);
      break;
    case WM_PAINT:
      BeginPaint(hWnd, &ps);
      DC = GetDC(hWnd);
      MemDC = CreateCompatibleDC(DC);
      bm = LoadBitmap(hInstance, MAKEINTRESOURCE(1));
      SelectObject(MemDC, bm);
      BitBlt(DC, 16, 16, 128, 96, MemDC, 0, 0, SRCCOPY);
      DeleteDC(MemDC);
      DeleteObject(bm);
      ReleaseDC(hWnd, DC);
      EndPaint(hWnd, &ps);
      break;

    case WM_DESTROY:
      EndDialog(hWnd, IDCANCEL);
  }
  return false;
}

void SetScrollBars(int width, int height, int x_pos, int y_pos, HWND hWnd)
{
  SCROLLINFO i;

  i.cbSize = sizeof(SCROLLINFO);
  i.fMask = SIF_ALL;
  i.nMin = 0;

  if (draw_buffer)
  {
    if (width < ts.width)
    {
      i.nMax = ts.width - 1;
      i.nPage = width;
      i.nPos = x_pos;
      SetScrollInfo(hWnd, SB_HORZ, &i, TRUE);
      ShowScrollBar(hWnd, SB_HORZ, TRUE);
    }
    else
      ShowScrollBar(hWnd, SB_HORZ, FALSE);
    if (height < ts.height)
    {
      i.nMax = ts.height - 1;
      i.nPage = height;
      i.nPos = y_pos;
      SetScrollInfo(hWnd, SB_VERT, &i, TRUE);
      ShowScrollBar(hWnd, SB_VERT, TRUE);
    }
    else
      ShowScrollBar(hWnd, SB_VERT, FALSE);
  }
  else
    ShowScrollBar(hWnd, SB_BOTH, FALSE);
}

void ConvertColor(long w1, long h1, long w2, long h2, long x, long y, int bps, char *buf)
{
  asm
  {
    mov esi, render_buf
    mov edi, buf
    xor edx, edx
    mov eax, [y]
    imul [w2]
    add eax, [x]
    mov ebx, eax
    shl eax, 1
    add eax, ebx
    add esi, eax
    mov ebx, [w2]
    sub ebx, [w1]
    mov eax, ebx
    shl ebx, 1
    add ebx, eax
    cmp [bps], 15
    je b16
    cmp [bps], 16
    je b16
    cmp [bps], 24
    je b24
    cmp [bps], 32
    je b32
    jmp exit
b16:mov ecx, [h1]
l5: push ecx
    mov ecx, [w1]
l6: xor edx, edx
    lodsb
    shr al, 3
    or dl, al
    ror edx, 5
    lodsb
    shr al, 3
    or dl, al
    ror edx, 5
    lodsb
    shr al, 3
    or dl, al
    rol edx, 10
    mov eax, edx
    stosw
    loop l6
    rep movsb
    pop ecx
    add esi, ebx
    loop l5
    jmp exit
b24:mov edx, [w1]
    mov eax, edx
    shl edx, 1
    add edx, eax
    and eax, 1
    mov ecx, [h1]
l1: push ecx
    mov ecx, edx
    rep movsb
    pop ecx
    add esi, ebx
    add edi, eax
    loop l1
    jmp exit
b32:mov edx, [w1]
    mov ecx, [h1]
l3: push ecx
    mov ecx, edx
l4: movsw
    movsb
    inc edi
    loop l4
    rep movsb
    pop ecx
    add esi, ebx
    loop l3
exit:
  }
}

LONG WINAPI wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  PAINTSTRUCT ps;
  HBITMAP Bitmap;
  HDC MemDC, DC;
  RECT r;
  char buf[256], *screen_buf;
  static int x_pos = 0, y_pos = 0;
  int w, h, b, c, bps, x, y;

  GetClientRect(hWnd, &r);

  switch(msg)
  {
    case WM_CREATE:
      CreateScene();
      return 0;

    case WM_PAINT:
      BeginPaint(hWnd, &ps);
      DC = GetDC(hWnd);
      SelectObject(DC, GetStockObject(BLACK_BRUSH));
      SelectObject(DC, GetStockObject(BLACK_PEN));
      if (draw_buffer)
      {
        w = r.right - r.left + 1;
        h = r.bottom - r.top + 1;
        if (w > ts.width)
        {
          x = (w - ts.width) >> 1;
          w = ts.width;
          x_pos = 0;
        }
        else
        {
          x = 0;
          x_pos = GetScrollPos(hWnd, SB_HORZ);
        }
        if (h > ts.height)
        {
          y = (h - ts.height) >> 1;
          h = ts.height;
          y_pos = 0;
        }
        else
        {
          y = 0;
          y_pos = GetScrollPos(hWnd, SB_VERT);
        }
        c = w & 1;
        bps = GetDeviceCaps(DC, BITSPIXEL);
        b = (bps + 1) >> 3;
        screen_buf = new char[w * h * b + h * c];
        ConvertColor(w, h, ts.width, ts.height, x_pos, y_pos, bps, screen_buf);
        MemDC = CreateCompatibleDC(DC);
        Bitmap = CreateBitmap(w, h, GetDeviceCaps(DC, PLANES), bps, NULL);
        SetBitmapBits(Bitmap, w * h * b + h * c, screen_buf);
        SelectObject(MemDC, Bitmap);
        BitBlt(DC, x, y, w, h, MemDC, 0, 0, SRCCOPY);
        DeleteDC(MemDC);
        DeleteObject(Bitmap);
        delete screen_buf;
        if (ts.width < r.right)
          Rectangle(DC, ts.width + x, 0, r.right, ts.height + y);
        if (ts.height < r.bottom)
          Rectangle(DC, x, ts.height + y, r.right, r.bottom);
        if (x)
          Rectangle(DC, 0, y, x, r.bottom);
        if (y)
          Rectangle(DC, 0, 0, x + ts.width, y);
      }
      else
        Rectangle(DC, r.left, r.top, r.right, r.bottom);
      ReleaseDC(hWnd, DC);
      EndPaint(hWnd, &ps);
      return 0;

    case WM_COMMAND:
      switch (LOWORD(wParam))
      {
        case CMD_RENDER:
          temp_struct = dlg_struct;
          if (DialogBox(hInstance, MAKEINTRESOURCE(1), hWnd, (DLGPROC)dlg1Proc) == IDCANCEL) break;
          if (dlg_struct.res != 9)
          {
            ts.width = widths[dlg_struct.res];
            ts.height = heights[dlg_struct.res];
          }
          else
          {
            ts.width = dlg_struct.x_res;
            ts.height = dlg_struct.y_res;
          }
          InvalidateRect(hWnd, NULL, false);
          DC = GetDC(hWnd);
          ts.bps = GetDeviceCaps(DC, BITSPIXEL);
          ReleaseDC(hWnd, DC);
          ts.hWnd = hWnd;
          CreateBuffers();
          RenderThread = CreateThread(NULL, 0, ThreadFunc, NULL, 0, &ret_val);
          draw_buffer = true;
          SetScrollBars(r.right - r.left, r.bottom - r.top, x_pos, y_pos, hWnd);
          x_pos = y_pos = 0;
          hDlg = CreateDialog(hInstance, MAKEINTRESOURCE(2), hWnd, (DLGPROC)dlg2Proc);
          break;

        case CMD_SAVEAS:
          strcpy(buf, dlg_struct.filename);
          if (GetFileName(buf, hWnd))
            SaveFile(buf);
          break;

        case CMD_EXIT:
          DestroyWindow(hWnd);
          break;

        case CMD_ABOUT:
          DialogBox(hInstance, MAKEINTRESOURCE(3), hWnd, (DLGPROC)dlg3Proc);
      }
      return 0;

    case MS_COMPLETED:
      if (dlg_struct.save) SaveFile(dlg_struct.filename);
      DestroyWindow(hDlg);
      EnableMenuItem(GetMenu(ts.hWnd), CMD_SAVEAS, MF_ENABLED);
      InvalidateRect(ts.hWnd, NULL, false);
      return 0;

    case WM_SIZE:
      SetScrollBars(LOWORD(lParam), HIWORD(lParam), x_pos, y_pos, hWnd);
      return 0;

    case WM_KEYDOWN:
      x_pos = GetScrollPos(hWnd, SB_HORZ);
      y_pos = GetScrollPos(hWnd, SB_VERT);
      switch (wParam)
      {
        case 38: y_pos--; break;
        case 40: y_pos++; break;
        case 37: x_pos--; break;
        case 39: x_pos++; break;
        case 33: y_pos -= r.bottom - r.top; break;
        case 34: y_pos += r.bottom - r.top; break;
        case 46: x_pos -= r.right - r.left; break;
        case 35: x_pos += r.right - r.left;
      }
      SetScrollPos(hWnd, SB_HORZ, x_pos, TRUE);
      SetScrollPos(hWnd, SB_VERT, y_pos, TRUE);
      InvalidateRect(hWnd, NULL, FALSE);
      break;

    case WM_HSCROLL:
      x_pos = GetScrollPos(hWnd, SB_HORZ);
      switch (LOWORD(wParam))
      {
        case SB_THUMBPOSITION:
        case SB_THUMBTRACK: x_pos = HIWORD(wParam); break;
        case SB_PAGELEFT: x_pos -= r.right - r.left + 1; break;
        case SB_LINELEFT: x_pos--; break;
        case SB_PAGERIGHT: x_pos += r.right - r.left + 1; break;
        case SB_LINERIGHT: x_pos++;
      }
      SetScrollPos(hWnd, SB_HORZ, x_pos, TRUE);
      InvalidateRect(hWnd, NULL, FALSE);
      return 0;

    case WM_VSCROLL:
      y_pos = GetScrollPos(hWnd, SB_VERT);
      switch (LOWORD(wParam))
      {
        case SB_THUMBPOSITION:
        case SB_THUMBTRACK: y_pos = HIWORD(wParam); break;
        case SB_PAGEDOWN: y_pos += r.bottom - r.top + 1; break;
        case SB_LINEDOWN: y_pos++; break;
        case SB_PAGEUP: y_pos -= r.bottom - r.top + 1; break;
        case SB_LINEUP: y_pos--;
      }
      SetScrollPos(hWnd, SB_VERT, y_pos, TRUE);
      InvalidateRect(hWnd, NULL, FALSE);
      return 0;

    case WM_DISPLAYCHANGE:
      if (wParam < 15)
      {
        draw_buffer = false;
        DestroyWindow(hDlg);
        ResetRenderer();
        MessageBox(hWnd, "The program doesn't support palette modes", "Error", MB_OK + MB_ICONSTOP);
        DestroyWindow(hWnd);
      }
      return 0;

    case WM_ERASEBKGND:
      return 1;

    case WM_DESTROY:
      ResetRenderer();
      delete scene;
      if (render_buf)
        delete render_buf;
      PostQuitMessage(0);
      return 0;
  }

  return DefWindowProc(hWnd, msg, wParam, lParam);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR cmdLine, int cmdShow)
{
  static char appName [] = "Ray tracing - by HackMaster";
  WNDCLASS wc;
  HWND hWnd;
  MSG msg;
  int width, height;

  ::hInstance = hInstance;
  HDC DC = CreateDC("Display", NULL, NULL, NULL);
  ts.bps = GetDeviceCaps(DC, BITSPIXEL);
  width = GetDeviceCaps(DC, HORZRES);
  height = GetDeviceCaps(DC, VERTRES);
  DeleteDC(DC);
  if (ts.bps <= 8)
  {
    MessageBox(0, "The program doesn't support palette modes", "Error", MB_OK + MB_ICONSTOP);
    return 0;
  }

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc =(WNDPROC)wndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(1));
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground =(HBRUSH)GetStockObject(BLACK_BRUSH);
  wc.lpszMenuName = MAKEINTRESOURCE(1);
  wc.lpszClassName = appName;

  RegisterClass(&wc);

  hWnd = CreateWindow(appName, appName,
                      WS_OVERLAPPEDWINDOW | WS_HSCROLL | WS_VSCROLL,
                      (width - WIDTH) >> 1, (height - HEIGHT) >> 1, WIDTH,
                      HEIGHT, HWND_DESKTOP, NULL, hInstance, NULL);

  ShowWindow(hWnd, cmdShow);
  UpdateWindow(hWnd);

  while(GetMessage(&msg, NULL, 0, 0))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }

  return msg.wParam;
}

